@@ -9,11 +9,8 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU General Public License (COPYING file) for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
  */
 #include <stdlib.h>
 #include <unistd.h>
@@ -38,6 +35,7 @@
 #include "process.h"
 #include "udp.h"
 #include "notify.h"
+#include "filter.h"
 
 #define FIRST_REPORT_SEC 3
 
@@ -50,9 +48,21 @@ static char *log_filename = NULL;
 static int local_syslog = 0;
 static int remote_syslog = 0;
 
+static int packet_from_file = 0;
+static int packet_buflen;
+static uint8_t packet_buf[256];
+static enum msg_type packet_type = ECM_MSG;
+
+extern int ai_family;
+
 static void do_log(FILE *f, time_t now, const char *msg) {
 	char date[64];
 	struct tm tm;
+	// There is no need to show timestamps when debug options are used
+	if (packet_from_file) {
+		fprintf(f, "%s", msg);
+		return;
+	}
 	localtime_r(&now, &tm);
 	strftime(date, sizeof(date), "%F %H:%M:%S", localtime_r(&now, &tm));
 	fprintf(f, "%s | %s", date, msg);
@@ -69,9 +79,9 @@ static void LOG_func(const char *msg) {
 		LOG(msg);
 }
 
-static const char short_options[] = "i:d:N:Sl:L:F:I:RzM:T:W:O:o:t:rk:g:pwxyc:C:Y:Q:A:s:U:P:B:eZ:Ef:X:H:G:KJ:D:jbhV";
+static const char short_options[] = "i:d:N:Sl:L:F:I:RzM:T:W:O:o:t:rk:g:upwxyc:C:Y:Q:A:s:U:P:B:46eZ:Ef:a:X:H:G:KJ:D:jbhVn:m:";
 
-// Unused short options: amnquv0123456789
+// Unused short options: aqv01235789
 static const struct option long_options[] = {
 	{ "ident",				required_argument, NULL, 'i' },
 	{ "daemon",				required_argument, NULL, 'd' },
@@ -94,7 +104,7 @@ static const struct option long_options[] = {
 	{ "output-rtp",			no_argument,       NULL, 'r' },
 	{ "output-rtp-ssrc",	required_argument, NULL, 'k' },
 	{ "output-tos",			required_argument, NULL, 'g' },
-	{ "output-filter",		no_argument,       NULL, 'p' },
+	{ "no-output-on-error",	no_argument,       NULL, 'u' },
 	{ "no-output-filter",	no_argument,       NULL, 'p' },
 	{ "output-nit-pass",	no_argument,       NULL, 'y' },
 	{ "output-eit-pass",	no_argument,       NULL, 'w' },
@@ -110,11 +120,14 @@ static const struct option long_options[] = {
 	{ "camd-user",			required_argument, NULL, 'U' },
 	{ "camd-pass",			required_argument, NULL, 'P' },
 	{ "camd-des-key",		required_argument, NULL, 'B' },
+	{ "ipv4",				no_argument,       NULL, '4' },
+	{ "ipv6",				no_argument,       NULL, '6' },
 
 	{ "emm",				no_argument,       NULL, 'e' },
 	{ "emm-pid",			required_argument, NULL, 'Z' },
 	{ "emm-only",			no_argument,       NULL, 'E' },
 	{ "emm-report-time",	required_argument, NULL, 'f' },
+	{ "emm-filter",			required_argument, NULL, 'a' },
 
 	{ "ecm-pid",			required_argument, NULL, 'X' },
 	{ "ecm-report-time",	required_argument, NULL, 'H' },
@@ -128,6 +141,9 @@ static const struct option long_options[] = {
 	{ "help",				no_argument,       NULL, 'h' },
 	{ "version",			no_argument,       NULL, 'V' },
 
+	{ "ecm-file",			required_argument, NULL, 'n' },
+	{ "emm-file",			required_argument, NULL, 'm' },
+
 	{ 0, 0, 0, 0 }
 };
 
@@ -144,9 +160,10 @@ static void show_help(struct ts *ts) {
 	printf("\n");
 	printf("Input options:\n");
 	printf(" -I --input <source>        | Where to read from. File or multicast address.\n");
-	printf("                            .    -I 224.0.0.1:5000 (multicast receive)\n");
-	printf("                            .    -I file.ts        (read from file)\n");
-	printf("                            .    -I -              (read from stdin) (default)\n");
+	printf("                            .    -I 224.0.0.1:5000    (v4 multicast)\n");
+	printf("                            .    -I [ff01::1111]:5000 (v6 multicast)\n");
+	printf("                            .    -I file://in.ts      (read from file)\n");
+	printf("                            . By default the input is stdin.\n");
 	printf(" -R --input-rtp             | Enable RTP input\n");
 	printf(" -z --input-ignore-disc     | Do not report discontinuty errors in input.\n");
 	printf(" -M --input-service <srvid> | Choose service id when input is MPTS.\n");
@@ -155,14 +172,18 @@ static void show_help(struct ts *ts) {
 	printf("\n");
 	printf("Output options:\n");
 	printf(" -O --output <dest>         | Where to send output. File or multicast address.\n");
-	printf("                            .    -O 239.0.0.1:5000 (multicast send)\n");
-	printf("                            .    -O file.ts        (write to file)\n");
-	printf("                            .    -O -              (write to stdout) (default)\n");
-	printf(" -o --output-intf <addr>    | Set multicast output interface. Default: %s\n", inet_ntoa(ts->output.intf));
+	printf("                            .    -O 239.0.0.1:5000    (v4 multicast)\n");
+	printf("                            .    -O [ff01::2222]:5000 (v6 multicast)\n");
+	printf("                            .    -O file://out.ts     (write to file)\n");
+	printf("                            . By default the output is stdout.\n");
+	printf(" -o --output-intf <value>   | Set multicast output interface.\n");
+	printf("                            . Default for IPv4: 0.0.0.0 (intf addr)\n");
+	printf("                            . Default for IPv6: -1      (intf number)\n");
 	printf(" -t --output-ttl <ttl>      | Set multicast ttl. Default: %d\n", ts->output.ttl);
 	printf(" -r --output-rtp            | Enable RTP output.\n");
 	printf(" -k --output-rtp-ssrc <id>  | Set RTP SSRC. Default: %u\n", ts->rtp_ssrc);
 	printf(" -g --output-tos <tos>      | Set TOS value of output packets. Default: none\n");
+	printf(" -u --no-output-on-error    | Do not output data when the code word is missing.\n");
 	printf(" -p --no-output-filter      | Disable output filtering. Default: %s\n", ts->pid_filter ? "enabled" : "disabled");
 	printf(" -y --output-nit-pass       | Pass through NIT.\n");
 	printf(" -w --output-eit-pass       | Pass through EIT (EPG).\n");
@@ -182,11 +203,19 @@ static void show_help(struct ts *ts) {
 	printf("CAMD server options:\n");
 	printf(" -A --camd-proto <proto>    | Set CAMD network protocol.\n");
 	printf("                            . Valid protocols are: CS378X (default) and NEWCAMD\n");
-	printf(" -s --camd-server <addr>    | Set CAMD server ip_address:port (1.2.3.4:2233).\n");
+	printf(" -s --camd-server <host>    | Set CAMD server address. Default port: 2233\n");
+	printf("                            . Example IPv4 addr and port: 127.0.0.1:2233\n");
+	printf("                            . Example IPv6 addr and port: [2a00::1014]:2233\n");
+	printf("                            . Example hostname          : example.com\n");
+	printf("                            . Example hostname and port : example.com:2233\n");
+	printf("                            . Example IPv4 hostname     : ipv4.google.com\n");
+	printf("                            . Example IPv6 hostname     : ipv6.google.com\n");
 	printf(" -U --camd-user <user>      | Set CAMD server user. Default: %s\n", ts->camd.user);
 	printf(" -P --camd-pass <pass>      | Set CAMD server password. Default: %s\n", ts->camd.pass);
 	printf(" -B --camd-des-key <key>    | Set DES key for newcamd protocol.\n");
 	printf("                            . Default: %s\n", ts->camd.newcamd.hex_des_key);
+	printf(" -4 --ipv4                  | Use only IPv4 addresses of the camd server.\n");
+	printf(" -6 --ipv6                  | Use only IPv6 addresses of the camd server.\n");
 	printf("\n");
 	printf("EMM options:\n");
 	printf(" -e --emm                   | Enable sending EMM's to CAMD. Default: %s\n", ts->emm_send ? "enabled" : "disabled");
@@ -196,6 +225,9 @@ static void show_help(struct ts *ts) {
 	printf(" -f --emm-report-time <sec> | Report each <sec> seconds how much EMMs have been\n");
 	printf("                            .   received/processed. Set <sec> to 0 to disable\n");
 	printf("                            .   the reports. Default: %d sec\n", ts->emm_report_interval);
+	printf(" -a --emm-filter <filter>   | Add EMM filter defined by <filter>.\n");
+	printf("                            . This option can be used multiple times (max:%u).\n", MAX_FILTERS);
+	printf("                            . See FILTERING file for more info.\n");
 	printf("\n");
 	printf("ECM options:\n");
 	printf(" -X --ecm-pid <pid>         | Force ECM pid. Default: none\n");
@@ -220,6 +252,10 @@ static void show_help(struct ts *ts) {
 	printf("                            .    4 = packet debug\n");
 	printf("                            .    5 = packet debug + packet dump\n");
 	printf("\n");
+	printf("Debugging options:\n");
+	printf(" -n --ecm-file <file.txt>   | Read ECM from text file.\n");
+	printf(" -m --emm-file <file.txt>   | Read EMM from text file.\n");
+	printf("\n");
 	printf("Misc options:\n");
 	printf(" -j --pid-report            | Report how much packets were received.\n");
 	printf(" -b --bench                 | Benchmark decrypton.\n");
@@ -229,32 +265,39 @@ static void show_help(struct ts *ts) {
 }
 
 static int parse_io_param(struct io *io, char *opt, int open_flags, mode_t open_mode) {
+	int port_set = 0, host_set;
 	io->type = WTF_IO;
-	char *p = strrchr(opt, ':');
-	if (!p) {
+	if (strstr(opt, "file://") == opt) {
+		io->fname = opt + 7; // strlen("file://")
+		io->type = FILE_IO;
+	} else if (strchr(opt, '/')) {
+		io->fname = opt;
 		io->type = FILE_IO;
-		if (strcmp(opt, "-") != 0) {
-			io->fd = open(opt, open_flags, open_mode);
-			if (io->fd < 0) {
-				fprintf(stderr, "ERROR: Can not open file (%s): %s\n", opt, strerror(errno));
-				exit(EXIT_FAILURE);
-			}
-		}
-		io->fname = strdup(opt);
-		return 0;
 	}
-	*p = 0x00;
-	io->type = NET_IO;
-	io->port = atoi(p + 1);
-	if (inet_aton(opt, &io->addr) == 0)
+	if (io->type == FILE_IO) {
+		io->fd = open(io->fname, open_flags, open_mode);
+		if (io->fd < 0) {
+			fprintf(stderr, "ERROR: Can not open file (%s): %s\n", io->fname, strerror(errno));
+			exit(EXIT_FAILURE);
+		}
 		return 1;
-	return 0;
+	}
+	io->type = NET_IO;
+	host_set = parse_host_and_port(opt, &io->hostname, &io->service, &port_set);
+	return !(!port_set || !host_set);
 }
 
+extern char *optarg;
+extern int optind, opterr, optopt;
+
 static void parse_options(struct ts *ts, int argc, char **argv) {
-	int j, i, ca_err = 0, server_err = 1, input_addr_err = 0, output_addr_err = 0, output_intf_err = 0, ident_err = 0, port_set = 0;
+	int j, i, ca_err = 0, server_err = 1, input_addr_err = 0, output_addr_err = 0, ident_err = 0, port_set = 0;
+	opterr = 0; // Prevent printing of error messages for unknown options in getopt()
 	while ((j = getopt_long(argc, argv, short_options, long_options, NULL)) != -1) {
-		char *p = NULL;
+		if (j == '?') {
+			fprintf(stderr, "ERROR: Unknown parameter '%s'.\n", argv[optind - 1]);
+			exit(EXIT_FAILURE);
+		}
 		switch (j) {
 			case 'i': // -- ident
 				ts->ident = optarg;
@@ -283,7 +326,7 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 				break;
 
 			case 'I': // --input
-				input_addr_err = parse_io_param(&ts->input, optarg, O_RDONLY, 0);
+				input_addr_err = !parse_io_param(&ts->input, optarg, O_RDONLY, 0);
 				break;
 			case 'R': // --input-rtp
 				ts->rtp_input = !ts->rtp_input;
@@ -302,13 +345,15 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 				break;
 
 			case 'O': // --output
-				output_addr_err = parse_io_param(&ts->output, optarg,
+				output_addr_err = !parse_io_param(&ts->output, optarg,
 					O_CREAT | O_WRONLY | O_TRUNC,
 					S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
 				break;
 			case 'o': // --output-intf
-				if (inet_aton(optarg, &ts->output.intf) == 0)
-					output_intf_err = 1;
+				if (strchr(optarg, '.'))
+					inet_aton(optarg, &ts->output.intf);
+				else
+					ts->output.v6_if_index = atoi(optarg);
 				break;
 			case 't': // --output-ttl
 				ts->output.ttl = atoi(optarg);
@@ -322,8 +367,11 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 			case 'g': // --output-tos
 				ts->output.tos = (uint8_t)strtol(optarg, NULL, 0);
 				break;
+			case 'u': // --no-output-on-error
+				ts->no_output_on_error = !ts->no_output_on_error;
+				break;
 			case 'p': // --no-output-filter
-				ts->pid_filter = 0;
+				ts->pid_filter = !ts->pid_filter;
 				break;
 			case 'y': // --output-nit-pass
 				ts->nit_passthrough = !ts->nit_passthrough;
@@ -418,16 +466,7 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 				}
 				break;
 			case 's': // --camd-server
-				p = strrchr(optarg, ':');
-				if (p) {
-					*p = 0x00;
-					ts->camd.server_port = atoi(p + 1);
-					port_set = 1;
-				}
-				if (inet_aton(optarg, &ts->camd.server_addr) == 0)
-					server_err = 1;
-				else
-					server_err = 0;
+				server_err = !parse_host_and_port(optarg, &ts->camd.hostname, &ts->camd.service, &port_set);
 				break;
 			case 'U': // --camd-user
 				if (strlen(optarg) < 64)
@@ -446,6 +485,12 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 				strncpy(ts->camd.newcamd.hex_des_key, optarg, sizeof(ts->camd.newcamd.hex_des_key) - 1);
 				ts->camd.newcamd.hex_des_key[sizeof(ts->camd.newcamd.hex_des_key) - 1] = 0;
 				break;
+			case '4': // --ipv4
+				ai_family = AF_INET;
+				break;
+			case '6': // --ipv6
+				ai_family = AF_INET6;
+				break;
 
 			case 'e': // --emm
 				ts->emm_send = !ts->emm_send;
@@ -463,6 +508,19 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 					ts->emm_report_interval = 86400;
 				break;
 
+			case 'a': // --emm-filter
+				if (ts->emm_filters_num + 1 > MAX_FILTERS) {
+					fprintf(stderr, "ERROR: Maximum allowed filters are %d.\n", MAX_FILTERS);
+					exit(EXIT_FAILURE);
+				}
+				if (filter_parse(optarg, &ts->emm_filters[ts->emm_filters_num])) {
+					ts->emm_filters_num++;
+				} else {
+					fprintf(stderr, "ERROR: Can't parse EMM filter: %s\n", optarg);
+					exit(EXIT_FAILURE);
+				}
+				break;
+
 			case 'X': // --ecm-pid
 				ts->forced_ecm_pid = strtoul(optarg, NULL, 0) & 0x1fff;
 				break;
@@ -496,6 +554,17 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 				csa_benchmark();
 				exit(EXIT_SUCCESS);
 
+			case 'n': // --ecm-file
+			case 'm': // --emm-file
+				packet_from_file = 1;
+				packet_buflen = file_hex2buf(optarg, packet_buf, sizeof(packet_buf));
+				if (!packet_buflen) {
+					fprintf(stderr, "ERROR: Can't init packet from file.\n");
+					exit(1);
+				}
+				packet_type = j == 'n' ? ECM_MSG : EMM_MSG;
+				break;
+
 			case 'h': // --help
 				show_help(ts);
 				exit(EXIT_SUCCESS);
@@ -510,6 +579,36 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 			ident_err = 1;
 	}
 
+	if (packet_from_file) {
+		int err = 0;
+		if (!ts->forced_caid) {
+			fprintf(stderr, "ERROR: CAID was not set. Use --caid option.\n");
+			err++;
+		}
+		if (!ts->forced_service_id) {
+			fprintf(stderr, "ERROR: Service id was not set. Use --input-service option.\n");
+			err++;
+		}
+		if (err)
+			exit(EXIT_FAILURE);
+
+		ts->threaded = 0;
+		input_addr_err = 0;
+		output_addr_err = 0;
+		ts->input.type = FILE_IO;
+		ts->input.fd = 0;
+		ts->output.type = FILE_IO;
+		ts->output.fd = 1;
+		ts->pid_filter = 0;
+		ts->emm_send = 1;
+		ts->emm_report_interval = 0;
+		ts->ecm_report_interval = 0;
+		ts->cw_warn_sec = 0;
+		ts->camd.no_reconnect = 1;
+		ts->camd.check_emm_errors = 1;
+		ts->emm_filters_num = 0;
+	}
+
 	// Constant codeword is special. Disable conflicting options
 	if (ts->camd.constant_codeword) {
 		server_err = 0; // No server settings are required
@@ -529,13 +628,11 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 		if (ca_err)
 			fprintf(stderr, "ERROR: Requested CA system is unsupported.\n");
 		if (server_err)
-			fprintf(stderr, "ERROR: CAMD server IP address is not set or it is invalid.\n");
+			fprintf(stderr, "ERROR: CAMD server address is not set or it is invalid.\n");
 		if (input_addr_err)
-			fprintf(stderr, "ERROR: Input IP address is invalid.\n");
+			fprintf(stderr, "ERROR: Input address is invalid.\n");
 		if (output_addr_err)
-			fprintf(stderr, "ERROR: Output IP address is invalid.\n");
-		if (output_intf_err)
-			fprintf(stderr, "ERROR: Output interface address is invalid.\n");
+			fprintf(stderr, "ERROR: Output address is invalid.\n");
 		exit(EXIT_FAILURE);
 	}
 	if (decode_hex_string(ts->camd.newcamd.hex_des_key, ts->camd.newcamd.bin_des_key, DESKEY_LENGTH) < 0) {
@@ -543,7 +640,7 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 		exit(EXIT_FAILURE);
 	}
 	if (ts->camd.ops.proto == CAMD_NEWCAMD && !port_set) {
-		fprintf(stderr, "ERROR: CAMD server port is not set. Use --camd-server %s:xxxx to set the port.\n", inet_ntoa(ts->camd.server_addr));
+		fprintf(stderr, "ERROR: CAMD server port is not set. Use --camd-server %s:xxxx to set the port.\n", ts->camd.hostname);
 		exit(EXIT_FAILURE);
 	}
 
@@ -566,8 +663,10 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 			ts_LOGf("Syslog     : %s:%d\n", ts->syslog_host, ts->syslog_port);
 		else
 			ts_LOGf("Syslog     : enabled\n");
-	} else
-		ts_LOGf("Syslog     : disabled\n");
+	} else {
+		if (!packet_from_file)
+			ts_LOGf("Syslog     : disabled\n");
+	}
 
 	if (!ts->camd.constant_codeword) {
 		if (ts->forced_caid)
@@ -587,14 +686,15 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 	}
 
 	if (ts->input.type == NET_IO) {
-		ts_LOGf("Input addr : %s://%s:%u/\n",
+		ts_LOGf("Input addr : %s://%s:%s/\n",
 			ts->rtp_input ? "rtp" : "udp",
-			inet_ntoa(ts->input.addr), ts->input.port);
+			ts->input.hostname, ts->input.service);
 		if (ts->input_buffer_time) {
 			ts_LOGf("Input buff : %u ms\n", ts->input_buffer_time);
 		}
 	} else if (ts->input.type == FILE_IO) {
-		ts_LOGf("Input file : %s\n", ts->input.fd == 0 ? "STDIN" : ts->input.fname);
+		if (!packet_from_file)
+			ts_LOGf("Input file : %s\n", ts->input.fd == 0 ? "STDIN" : ts->input.fname);
 	}
 	if (ts->input_dump_filename) {
 		ts->input_dump_file = fopen(ts->input_dump_filename, "w");
@@ -612,10 +712,11 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 	if (!ts->emm_only)
 	{
 		if (ts->output.type == NET_IO) {
-			ts_LOGf("Output addr: %s://%s:%u/\n",
+			ts_LOGf("Output addr: %s://%s:%s/\n",
 				ts->rtp_output ? "rtp" : "udp",
-				inet_ntoa(ts->output.addr), ts->output.port);
-			ts_LOGf("Output intf: %s\n", inet_ntoa(ts->output.intf));
+				ts->output.hostname, ts->output.service);
+			ts_LOGf("Output intf: %s (IPv6 intf index:%d)\n",
+				inet_ntoa(ts->output.intf), ts->output.v6_if_index);
 			ts_LOGf("Output ttl : %d\n", ts->output.ttl);
 			if (ts->output.tos > -1)
 				ts_LOGf("Output TOS : %u (0x%02x)\n", ts->output.tos, ts->output.tos);
@@ -626,12 +727,16 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 				RAND_bytes((unsigned char *)&(ts->rtp_seqnum), 2);
 			}
 		} else if (ts->output.type == FILE_IO) {
-			ts_LOGf("Output file: %s\n", ts->output.fd == 1 ? "STDOUT" : ts->output.fname);
+			if (!packet_from_file)
+				ts_LOGf("Output file: %s\n", ts->output.fd == 1 ? "STDOUT" : ts->output.fname);
+		}
+		if (!packet_from_file) {
+			ts_LOGf("Out filter : %s (%s)%s\n",
+				ts->pid_filter ? "enabled" : "disabled",
+				ts->pid_filter ? "output only service related PIDs" : "output everything",
+				ts->no_output_on_error ? " (No output on CW error)" : ""
+			);
 		}
-		ts_LOGf("Out filter : %s (%s)\n",
-			ts->pid_filter ? "enabled" : "disabled",
-			ts->pid_filter ? "output only service related PIDs" : "output everything"
-		);
 		if (ts->pid_filter) {
 			if (ts->nit_passthrough)
 				ts_LOGf("Out filter : Pass through NIT.\n");
@@ -643,43 +748,68 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 	}
 	if (!ts->camd.constant_codeword) {
 		ts_LOGf("CAMD proto : %s\n", ts->camd.ops.ident);
-		ts_LOGf("CAMD addr  : tcp://%s:%u/\n", inet_ntoa(ts->camd.server_addr), ts->camd.server_port);
+		ts_LOGf("CAMD addr  : %s:%s%s\n", ts->camd.hostname, ts->camd.service,
+			ai_family == AF_INET  ? " (IPv4 only)" :
+			ai_family == AF_INET6 ? " (IPv6 only)" :
+			" (IPv4/IPv6)"
+		);
 		ts_LOGf("CAMD user  : %s\n", ts->camd.user);
 		ts_LOGf("CAMD pass  : %s\n", ts->camd.pass);
 		if (ts->camd.ops.proto == CAMD_NEWCAMD)
 			ts_LOGf("CAMD deskey: %s\n", ts->camd.newcamd.hex_des_key);
 	}
 
-	ts_LOGf("TS discont : %s\n", ts->ts_discont ? "report" : "ignore");
-	ts->threaded = !(ts->input.type == FILE_IO && ts->input.fd != 0);
-	if (ts->emm_send && ts->emm_report_interval)
-		ts_LOGf("EMM report : %d sec\n", ts->emm_report_interval);
-	if (ts->emm_send && ts->emm_report_interval == 0)
-		ts_LOGf("EMM report : disabled\n");
-	if (ts->forced_emm_pid)
-		ts_LOGf("EMM pid    : 0x%04x (%d)\n", ts->forced_emm_pid, ts->forced_emm_pid);
-	if (ts->emm_only) {
-		ts_LOGf("EMM only   : %s\n", ts->emm_only ? "yes" : "no");
-	} else {
-		if (!ts->camd.constant_codeword)
-			ts_LOGf("EMM send   : %s\n", ts->emm_send   ? "enabled" : "disabled");
+	if (!packet_from_file) {
+		ts_LOGf("TS discont : %s\n", ts->ts_discont ? "report" : "ignore");
+		ts->threaded = !(ts->input.type == FILE_IO && ts->input.fd != 0);
+	}
+
+	if (!packet_from_file && !ts->emm_only) {
 		ts_LOGf("Decoding   : %s\n", ts->threaded ? "threaded" : "single thread");
 	}
 
-	if (!ts->emm_only && ts->ecm_report_interval)
-		ts_LOGf("ECM report : %d sec\n", ts->emm_report_interval);
-	if (!ts->emm_only && ts->ecm_report_interval == 0 && !ts->camd.constant_codeword)
-		ts_LOGf("ECM report : disabled\n");
-	if (ts->forced_ecm_pid)
-		ts_LOGf("ECM pid    : 0x%04x (%d)\n", ts->forced_ecm_pid, ts->forced_ecm_pid);
+	if (!packet_from_file) {
+		if (ts->emm_send && ts->emm_report_interval)
+			ts_LOGf("EMM report : %d sec\n", ts->emm_report_interval);
+		if (ts->emm_send && ts->emm_report_interval == 0)
+			ts_LOGf("EMM report : disabled\n");
+		if (ts->forced_emm_pid)
+			ts_LOGf("EMM pid    : 0x%04x (%d)\n", ts->forced_emm_pid, ts->forced_emm_pid);
+	}
+	if (ts->emm_only) {
+		ts_LOGf("EMM only   : %s\n", ts->emm_only ? "yes" : "no");
+	} else {
+		if (!packet_from_file) {
+			if (!ts->camd.constant_codeword)
+				ts_LOGf("EMM send   : %s\n", ts->emm_send   ? "enabled" : "disabled");
+			ts_LOGf("Decoding   : %s\n", ts->threaded ? "threaded" : "single thread");
+		}
+	}
 
-	if (!ts->emm_only && ts->cw_warn_sec)
-		ts_LOGf("CW warning : %d sec\n", ts->cw_warn_sec);
-	if (!ts->emm_only && ts->cw_warn_sec && !ts->camd.constant_codeword)
-		ts_LOGf("CW warning : disabled\n");
+	if (ts->emm_send) {
+		for (i = 0; i < ts->emm_filters_num; i++) {
+			char tmp[512];
+			filter_dump(&ts->emm_filters[i], tmp, sizeof(tmp));
+			ts_LOGf("EMM filter : [%2d] %s\n", i + 1,  tmp);
+		}
+	}
 
-	if (!ts->ecm_cw_log && !ts->camd.constant_codeword)
-		ts_LOGf("ECM/CW log : disabled\n");
+	if (!packet_from_file) {
+		if (!ts->emm_only && ts->ecm_report_interval)
+			ts_LOGf("ECM report : %d sec\n", ts->emm_report_interval);
+		if (!ts->emm_only && ts->ecm_report_interval == 0 && !ts->camd.constant_codeword)
+			ts_LOGf("ECM report : disabled\n");
+		if (ts->forced_ecm_pid)
+			ts_LOGf("ECM pid    : 0x%04x (%d)\n", ts->forced_ecm_pid, ts->forced_ecm_pid);
+
+		if (!ts->emm_only && ts->cw_warn_sec)
+			ts_LOGf("CW warning : %d sec\n", ts->cw_warn_sec);
+		if (!ts->emm_only && ts->cw_warn_sec && !ts->camd.constant_codeword)
+			ts_LOGf("CW warning : disabled\n");
+
+		if (!ts->ecm_cw_log && !ts->camd.constant_codeword)
+			ts_LOGf("ECM/CW log : disabled\n");
+	}
 
 	if (ts->ident) {
 		int len = strlen(ts->ident);
@@ -691,7 +821,9 @@ static void parse_options(struct ts *ts, int argc, char **argv) {
 }
 
 static void report_emms(struct ts *ts, time_t now) {
-	ts_LOGf("EMM | Received %u and processed %u in %lu seconds.\n",
+	ts_LOGf("EMM | Received %u, Skipped %u, Sent %u, Processed %u in %lu seconds.\n",
+		ts->emm_input_count,
+		ts->emm_skipped_count,
 		ts->emm_seen_count,
 		ts->emm_processed_count,
 		now - ts->emm_last_report);
@@ -700,7 +832,9 @@ static void report_emms(struct ts *ts, time_t now) {
 			now - ts->emm_last_report);
 	}
 	ts->emm_last_report = now;
+	ts->emm_input_count = 0;
 	ts->emm_seen_count = 0;
+	ts->emm_skipped_count = 0;
 	ts->emm_processed_count = 0;
 }
 
@@ -832,6 +966,16 @@ int main(int argc, char **argv) {
 	ts.emm_last_report = time(NULL) + FIRST_REPORT_SEC;
 	ts.ecm_last_report = time(NULL) + FIRST_REPORT_SEC;
 	camd_start(&ts);
+
+	if (packet_from_file) {
+		uint8_t tmp[2048];
+		ts_hex_dump_buf((char *)tmp, sizeof(tmp), packet_buf, packet_buflen, 16);
+		ts_LOGf("%s | Processing packet with CAID 0x%04x\n", packet_type == ECM_MSG ? "ECM" : "EMM", ts.forced_caid);
+		ts_LOGf("%s | Packet dump:\n%s\n", packet_type == ECM_MSG ? "ECM" : "EMM", tmp);
+		camd_process_packet(&ts, camd_msg_alloc(packet_type, ts.forced_caid, ts.forced_service_id, packet_buf, packet_buflen));
+		goto EXIT;
+	}
+
 	do {
 		if (!ts.camd.constant_codeword)
 			do_reports(&ts);
@@ -861,16 +1005,16 @@ int main(int argc, char **argv) {
 				ts_LOGf("--- | Input read timeout.\n");
 				if (!ntimeouts) {
 					timeout_start = time(NULL);
-					notify(&ts, "INPUT_TIMEOUT", "Read timeout on input %s://%s:%u/",
+					notify(&ts, "INPUT_TIMEOUT", "Read timeout on input %s://%s:%s/",
 							ts.rtp_input ? "rtp" : "udp",
-							inet_ntoa(ts.input.addr), ts.input.port);
+							ts.input.hostname, ts.input.service);
 				}
 				ntimeouts++;
 			} else {
 				if (ntimeouts && readen > 0) {
-					notify(&ts, "INPUT_OK", "Data is available on input %s://%s:%u/ after %ld seconds timeout.",
+					notify(&ts, "INPUT_OK", "Data is available on input %s://%s:%s/ after %ld seconds timeout.",
 							ts.rtp_input ? "rtp" : "udp",
-							inet_ntoa(ts.input.addr), ts.input.port,
+							ts.input.hostname, ts.input.service,
 							(time(NULL) - timeout_start) + 2); // Timeout is detected when ~2 seconds there is no incoming data
 					ntimeouts = 0;
 				}
